import os

from abejacli.config import (
    ABEJA_API_URL,
    PLATFORM_REQUEST_TIMEOUT_SECONDS,
)
from abejacli.session import generate_user_session
from abejacli.config import DATALAKE_ITEMS_PER_PAGE
from abejacli.logger import get_logger
from abejacli.datalake.process_file_job import process_file_jobs
from abejacli.datalake.upload_job import upload_job
from abejacli.datalake.download_job import download_job
from abejacli.exceptions import InvalidDatalakeTimeInterval


logger = get_logger()


def generate_channel_file_iter_by_period(channel_id, start=None, end=None):
    """
    generate file iterator in channel from specified start date to specified end date

    :param channel_id: datalake channel identifier
    :param start: start date (YYYYMMDD)
    :param end: send date (YYYYMMDD)
    :return:
    """
    url = "{}/channels/{}".format(ABEJA_API_URL, channel_id)
    params = {
        'items_per_page': DATALAKE_ITEMS_PER_PAGE
    }

    if start and end:
        params['start'] = start
        params['end'] = end
    if (start and not end) or (not start and end):
        logger.error('both start and end are required for period of datalake file list')
        raise InvalidDatalakeTimeInterval()

    # Get upload url
    session = generate_user_session()

    while True:
        r = session.get(url, params=params, timeout=PLATFORM_REQUEST_TIMEOUT_SECONDS)
        r.raise_for_status()
        res = r.json()
        files = res.get('files')
        if not files or len(files) == 0:
            break
        # Iterate files
        for file_info in files:
            yield file_info
        next_page_token = res.get('next_page_token')
        if not next_page_token:
            break
        params = {
            'next_page_token': next_page_token
        }


def generate_channel_file_iter_by_id(channel_id, *file_ids):
    """
    generate file iterator for list of datalake file identifiers

    :param channel_id: datalake channel identifier
    :param file_ids: list of file id to iterate
    :return:
    """
    session = generate_user_session()

    for file_id in file_ids:
        url = "{}/channels/{}/{}".format(ABEJA_API_URL, channel_id, file_id)
        r = session.get(url, timeout=PLATFORM_REQUEST_TIMEOUT_SECONDS)
        r.raise_for_status()
        file_info = r.json()
        yield file_info


def download_from_datalake(channel_id, file_iter, target_dir):
    """
    download files and store into target dier

    :param channel_id: channel identifier
    :param file_iter: iterator of file to download
    :param target_dir: download target dir
    :return:
    """
    file_list = list(file_iter)
    file_num = len(file_list)

    # Setup worker_option
    worker_option = {
        'download_dir': target_dir
    }
    return process_file_jobs(channel_id, download_job, file_list, 'counter', file_num, worker_option)


def upload_to_datalake(channel_id, path_iter, options):
    """
    upload files in path iterator to datalake channel

    :param channel_id: channel identifier
    :param path_iter: iterator of paths to upload
    :return:
    """
    file_list = []
    total_size = 0
    for file_path in path_iter:
        total_size = total_size + os.path.getsize(file_path)
        file_list.append(file_path)

    return process_file_jobs(channel_id, upload_job, file_list, 'size', total_size, options)
