import os
import json
import socket
import tempfile

import requests
import docker
from docker.errors import ImageNotFound
from abejacli.model import md5file


DOCKERFILE_RUN_LOCAL_TEMPLATE = '''
FROM {IMAGE}

ADD . /tmp/app
WORKDIR /tmp/app
ENV SERVICE_TYPE {SERVICE_TYPE}
ENV HANDLER {HANDLER}
RUN if test -r requirements.txt; then pip install --no-cache-dir -r requirements.txt; fi
'''

DOCKERFILE_TRAIN_LOCAL_TEMPLATE = '''
FROM {IMAGE}

ADD . /tmp/app
WORKDIR /tmp/app
ENV DATASETS {DATASETS}
ENV HANDLER {HANDLER}
ENV ABEJA_PLATFORM_USER_ID {ABEJA_PLATFORM_USER_ID}
ENV ABEJA_PLATFORM_PERSONAL_ACCESS_TOKEN {ABEJA_PLATFORM_PERSONAL_ACCESS_TOKEN}
ENV ABEJA_ORGANIZATION_ID {ABEJA_ORGANIZATION_ID}
RUN if test -r requirements.txt; then pip install --no-cache-dir -r requirements.txt; fi
'''

REQUIREMENTS_TXT = 'requirements.txt'
BUILT_IMAGE_SUFFIX = 'local-model'
LOCAL_MODEL_TYPE_KEY = 'abeja-platform-model-type'
LOCAL_MODEL_REQUIREMENTS_MD5_KEY = 'abeja-platform-requirement-md5'

LOCAL_MODEL_TYPE_VALUE = 'inference'
LOCAL_TRAIN_TYPE_VALUE = 'train'


class LocalModelHandler:
    def __init__(self):
        self._client = docker.from_env()
        self._setup_docker_cli()
        self.requirements_file = REQUIREMENTS_TXT

    def _setup_docker_cli(self):
        """
        setup cli

        docker client will default to connecting to the following respectively.
        unix://var/run/docker.sock : linux / osx
        tcp://127.0.0.1:2376       : windows

        judge operation system using by os.name
        cf. https://docs.python.org/3.5/library/os.html#os.name

        :return:
        """
        if os.name == 'nt':
            base_url = 'tcp://127.0.0.1:2376'
        else:
            base_url = 'unix://var/run/docker.sock'
        self._docker_cli = docker.APIClient(base_url=base_url)

    def check_installation(self):
        try:
            return self._client.version()
        except requests.exceptions.ConnectionError:
            return None

    def pull_image(self, image, tag, stdout=None):
        """
        pull image if the image does not exist

        :param image:
        :param tag:
        :param stdout:
        :return:
        """
        if self._get_image(image, tag):
            return

        for output in self._docker_cli.pull(image, tag=tag, stream=True):
            if not stdout:
                continue
            try:
                for line in self._parse_stream(output):
                    line = line['status'].rstrip()
                    stdout(str(line))
            except Exception:
                continue

    def check_rebuild_required(self):
        """
        return true if requirements.txt exists and
        the last update is later than that of the specified image.
        target_last_modified and image_last_modified are time in utc.

        :param image:
        :param tag:
        :return:
        """
        # no need to rebuild if requirements.txt does not exist
        if not os.path.exists(self.requirements_file):
            return False

        image = self._find_built_image()
        if image is None:
            return True

        # check md5 of requirements.txt using label
        prev_md5 = image.labels.get(LOCAL_MODEL_REQUIREMENTS_MD5_KEY)
        if not prev_md5:
            return True

        if prev_md5 != md5file(self.requirements_file):
            return True

        return False

    def build_run_image(self,
                        image,
                        tag,
                        handler,
                        env_vars=None,
                        no_cache=False,
                        stdout=None):
        """
        build docker image adding label to be able to find it

        :param image:
        :param tag:
        :param handler:
        :param no_cache:
        :param stdout:
        :return:
        """
        if not no_cache:
            # when rebuild is required, don't use cache
            no_cache = self.check_rebuild_required()

        # skip build if use cache is enabled and built image exists.
        if self._find_built_image() and not no_cache:
            return

        if env_vars is None:
            env_vars = {}

        dockerfile = self._generate_run_dockerfile(image, tag, handler, env_vars)

        # ex. abeja/platform-minimal/0.1.0/local-inference-model
        image_name = '{}/{}/{}-{}'.format(image, tag, LOCAL_MODEL_TYPE_VALUE, BUILT_IMAGE_SUFFIX)

        self._build_image(image_name, dockerfile, LOCAL_MODEL_TYPE_VALUE, stdout)

    def build_train_image(self,
                          image,
                          tag,
                          handler,
                          organization_id,
                          user_id,
                          access_token,
                          datasets,
                          env_vars=None,
                          no_cache=False,
                          stdout=None):
        """
        build docker image adding label to be able to find it

        :param image:
        :param tag:
        :param handler:
        :param organization_id:
        :param user_id:
        :param access_token
        :param datasets:
        :param env_vars:
        :param no_cache:
        :param stdout:
        :return:
        """
        if not no_cache:
            # when rebuild is required, don't use cache
            no_cache = self.check_rebuild_required()

        # skip build if use cache is enabled and built image exists.
        if self._find_built_image() and not no_cache:
            return

        if env_vars is None:
            env_vars = {}

        dockerfile = self._generate_train_dockerfile(image, tag, handler,
                                                     datasets, user_id,
                                                     access_token, organization_id,
                                                     env_vars)

        # ex. abeja/platform-minimal/0.1.0/local-model
        image_name = '{}/{}/{}-{}'.format(image, tag, LOCAL_TRAIN_TYPE_VALUE, BUILT_IMAGE_SUFFIX)

        self._build_image(image_name, dockerfile, LOCAL_TRAIN_TYPE_VALUE, stdout)

    def _build_image(self, name, dockerfile, type_value=LOCAL_MODEL_TYPE_VALUE, stdout=None):
        with tempfile.NamedTemporaryFile(mode='w+t') as f:
            f.write(dockerfile)
            f.seek(0)  # put file pointer back the initial position for being read
            requirements_md5 = md5file(self.requirements_file) if os.path.exists(self.requirements_file) else None
            labels = {
                LOCAL_MODEL_TYPE_KEY: type_value,
                LOCAL_MODEL_REQUIREMENTS_MD5_KEY: requirements_md5
            }
            for output in self._docker_cli.build(tag=name,
                                                 dockerfile=f.name,
                                                 path=os.getcwd(),
                                                 labels=labels):
                if not stdout:
                    continue
                try:
                    for line in self._parse_stream(output):
                        line = line['stream'].rstrip()
                        stdout(str(line))
                except Exception:
                    continue

    def run_image(self):
        """
        run container
        :return:
        """
        image_id = self._find_built_image_id()
        container = self._client.containers.run(image_id,
                                                remove=True,
                                                detach=True,
                                                stderr=True,
                                                working_dir='/tmp/app',
                                                environment=[
                                                    'PYTHONUNBUFFERED=x'
                                                ],
                                                volumes={'{}'.format(os.getcwd()): {'bind': '/tmp/app', 'mode': 'rw'}},
                                                command=['/bin/sh', '-c', 'abeja-model train'])
        return container

    def stop_image(self, container):
        container.stop()

    def create_local_server(self):
        """
        run container in background as local api server
        assign dynamic port number for host port.

        :return:
        """
        image_id = self._find_built_image_id()
        random_port = self._get_free_tcp_port()
        container = self._client.containers.run(image_id,
                                                remove=True,
                                                detach=True,
                                                stderr=True,
                                                working_dir='/tmp/app',
                                                environment=[
                                                    'PYTHONUNBUFFERED=x'
                                                ],
                                                ports={'5000/tcp': random_port},
                                                volumes={'{}'.format(os.getcwd()): {'bind': '/tmp/app', 'mode': 'rw'}},
                                                command=['/bin/sh', '-c', 'abeja-model run'])

        return LocalServer(container, port=random_port)

    def _get_image(self, image, tag):
        try:
            return self._client.images.get("{}:{}".format(image, tag))
        except ImageNotFound:
            return None

    def _find_built_image(self):
        images = self._client.images.list(filters={'label': LOCAL_MODEL_TYPE_KEY})
        if not images:
            return None
        # pick the first because images is descending order list
        return images[0]

    def _find_built_image_id(self):
        image = self._find_built_image()
        _, image_id = image.short_id.split(':')
        return image_id

    def _generate_run_dockerfile(self, image, tag, handler, env_vars):
        image = "{}:{}".format(image, tag)
        dockerfile = DOCKERFILE_RUN_LOCAL_TEMPLATE.format(
            IMAGE=image, SERVICE_TYPE='HTTP', HANDLER=str(handler))
        for k, v in env_vars.items():
            dockerfile += '\nENV {} {}'.format(k, v)
        return dockerfile

    def _generate_train_dockerfile(self,
                                   image,
                                   tag,
                                   handler,
                                   datasets,
                                   user_id,
                                   access_token,
                                   organization_id,
                                   env_vars):
        image = "{}:{}".format(image, tag)
        dockerfile = DOCKERFILE_TRAIN_LOCAL_TEMPLATE.format(
            IMAGE=image, HANDLER=str(handler), DATASETS="'" + json.dumps(datasets) + "'",
            ABEJA_PLATFORM_USER_ID=str(user_id),
            ABEJA_PLATFORM_PERSONAL_ACCESS_TOKEN=str(access_token),
            ABEJA_ORGANIZATION_ID=str(organization_id))
        for k, v in env_vars.items():
            dockerfile += '\nENV {} {}'.format(k, v)
        return dockerfile

    def _parse_stream(self, output):
        if type(output) == bytes:
            # for py3
            output = output.decode('utf-8')
        lines = output.rstrip().split('\r\n')
        for line in lines:
            yield json.loads(line)

    def _get_free_tcp_port(self):
        """
        get free tcp port in dynamic port range
        which is defined by IANA
        cf. https://www.iana.org/assignments/service-names-port-numbers/service-names-port-numbers.xhtml

        :return:
        """
        tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        tcp.bind(('', 0))
        addr, port = tcp.getsockname()
        tcp.close()
        return port


class LocalServer:
    def __init__(self, container, port=None):
        self._container = container
        self.endpoint = 'http://localhost:{}'.format(port)

    def stop(self):
        # much faster to use kill, because stop container takes about 10 sec.
        # self._container.stop()
        try:
            self._container.kill()
        except docker.errors.NotFound:
            # ignore because the container does not exist anymore.
            return

    def logs(self, follow=True, stream=True, since=None):
        return self._container.logs(follow=follow, stream=stream, since=since)
